import ast

from boa_restrictor.exceptions.syntax_errors import BoaRestrictorParsingError


def parse_source_code_or_fail(*, filename: str, source_code: str) -> ast.Module:
    """
    Parse code through abstract syntax tree.
    Raise a human-readable error message if parsing fails.
    """
    try:
        return ast.parse(source_code)
    except SyntaxError as e:
        raise BoaRestrictorParsingError(filename=filename) from e



import ast
from pathlib import Path

from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules import NoLoopsInTestsRule


def test_pytest_for_loop_is_detected():
    source_tree = ast.parse("""def test_something():
    for i in list:
        pass""")

    occurrences = NoLoopsInTestsRule.run_check(file_path=Path("/path/to/tests/test_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="test_file.py",
        file_path=Path("/path/to/tests/test_file.py"),
        line_number=1,
        rule_id=NoLoopsInTestsRule.RULE_ID,
        rule_label=NoLoopsInTestsRule.RULE_LABEL,
        identifier="test_something",
    )


def test_pytest_while_loop_is_detected():
    source_tree = ast.parse("""def test_something():
    while i < 0:
        pass""")

    occurrences = NoLoopsInTestsRule.run_check(file_path=Path("/path/to/tests/test_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="test_file.py",
        file_path=Path("/path/to/tests/test_file.py"),
        line_number=1,
        rule_id=NoLoopsInTestsRule.RULE_ID,
        rule_label=NoLoopsInTestsRule.RULE_LABEL,
        identifier="test_something",
    )


def test_pytest_list_comprehension_loop_is_detected():
    source_tree = ast.parse("""def test_something():
    numbers = [1, 2, 3, 4, 5]
    squared_numbers = [x * x for x in numbers]""")

    occurrences = NoLoopsInTestsRule.run_check(file_path=Path("/path/to/tests/test_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="test_file.py",
        file_path=Path("/path/to/tests/test_file.py"),
        line_number=1,
        rule_id=NoLoopsInTestsRule.RULE_ID,
        rule_label=NoLoopsInTestsRule.RULE_LABEL,
        identifier="test_something",
    )


def test_pytest_set_comprehension_loop_is_detected():
    source_tree = ast.parse("""def test_something():
    {x ** 2 for x in range(10) if x % 2 == 0}""")

    occurrences = NoLoopsInTestsRule.run_check(file_path=Path("/path/to/tests/test_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="test_file.py",
        file_path=Path("/path/to/tests/test_file.py"),
        line_number=1,
        rule_id=NoLoopsInTestsRule.RULE_ID,
        rule_label=NoLoopsInTestsRule.RULE_LABEL,
        identifier="test_something",
    )


def test_pytest_dict_comprehension_loop_is_detected():
    source_tree = ast.parse("""def test_something():
    {x: x ** 2 for x in range(5)}""")

    occurrences = NoLoopsInTestsRule.run_check(file_path=Path("/path/to/tests/test_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="test_file.py",
        file_path=Path("/path/to/tests/test_file.py"),
        line_number=1,
        rule_id=NoLoopsInTestsRule.RULE_ID,
        rule_label=NoLoopsInTestsRule.RULE_LABEL,
        identifier="test_something",
    )


def test_pytest_nested_loop_is_detected():
    source_tree = ast.parse("""def test_something():
    if len(list) > 0:
        for i in list:
            pass""")

    occurrences = NoLoopsInTestsRule.run_check(file_path=Path("/path/to/tests/test_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="test_file.py",
        file_path=Path("/path/to/tests/test_file.py"),
        line_number=1,
        rule_id=NoLoopsInTestsRule.RULE_ID,
        rule_label=NoLoopsInTestsRule.RULE_LABEL,
        identifier="test_something",
    )


def test_pytest_no_loop_is_detected():
    source_tree = ast.parse("""def test_something():
    if len(list) > 0:
        pass""")

    occurrences = NoLoopsInTestsRule.run_check(file_path=Path("/path/to/tests/test_file.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_unittest_for_loop_is_detected():
    source_tree = ast.parse("""class MyTestCase(TestCase):
    def test_something(self):
        for i in list:
            pass""")

    occurrences = NoLoopsInTestsRule.run_check(file_path=Path("/path/to/tests/test_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="test_file.py",
        file_path=Path("/path/to/tests/test_file.py"),
        line_number=2,
        rule_id=NoLoopsInTestsRule.RULE_ID,
        rule_label=NoLoopsInTestsRule.RULE_LABEL,
        identifier="test_something",
    )


def test_unittest_while_loop_is_detected():
    source_tree = ast.parse("""class MyTestCase(TestCase):
    def test_something():
        while i < 0:
            pass""")

    occurrences = NoLoopsInTestsRule.run_check(file_path=Path("/path/to/tests/test_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="test_file.py",
        file_path=Path("/path/to/tests/test_file.py"),
        line_number=2,
        rule_id=NoLoopsInTestsRule.RULE_ID,
        rule_label=NoLoopsInTestsRule.RULE_LABEL,
        identifier="test_something",
    )


def test_unittest_loop_is_detected_from_non_test_case_class():
    source_tree = ast.parse("""class MyTestCase(InheritedTestCase):
    def test_something():
        while i < 0:
            pass""")

    occurrences = NoLoopsInTestsRule.run_check(file_path=Path("/path/to/tests/test_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="test_file.py",
        file_path=Path("/path/to/tests/test_file.py"),
        line_number=2,
        rule_id=NoLoopsInTestsRule.RULE_ID,
        rule_label=NoLoopsInTestsRule.RULE_LABEL,
        identifier="test_something",
    )


def test_loops_ok_in_non_test_dir():
    source_tree = ast.parse("""def test_something():
    while i < 0:
        pass""")

    occurrences = NoLoopsInTestsRule.run_check(file_path=Path("/path/to/code/test_file.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_loops_ok_in_non_test_file():
    source_tree = ast.parse("""def test_something():
    while i < 0:
        pass""")

    occurrences = NoLoopsInTestsRule.run_check(file_path=Path("/path/to/tests/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_is_test_file_test_file_in_test_dir():
    assert NoLoopsInTestsRule._is_test_file(filepath=Path("path/to/tests/test_file.py")) is True


def test_is_test_file_test_file_but_not_test_dir():
    assert NoLoopsInTestsRule._is_test_file(filepath=Path("path/to/code/test_file.py")) is False


def test_is_test_file_no_test_file_in_test_dir():
    assert NoLoopsInTestsRule._is_test_file(filepath=Path("path/to/tests/mixins.py")) is False


def test_is_test_file_test_file_in_test_dir_but_not_py():
    assert NoLoopsInTestsRule._is_test_file(filepath=Path("path/to/tests/test_file.md")) is False



import ast
from pathlib import Path

from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules.django.prohibit_assert_raises import AssertRaisesProhibitedRule


def test_check_occurrence_duplication_no_duplication():
    rule = AssertRaisesProhibitedRule(file_path=Path("/path/to/file/my_file.py"), source_tree=ast.parse("a=1"))

    assert rule._check_occurrence_duplication(occurrences=[], filename="my_file.py", line_number=7) is False


def test_check_occurrence_duplication_no_duplication_via_file():
    rule = AssertRaisesProhibitedRule(file_path=Path("/path/to/file/my_file.py"), source_tree=ast.parse("a=1"))
    occurrence = Occurrence(
        line_number=7,
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        rule_id=AssertRaisesProhibitedRule.RULE_ID,
        rule_label=AssertRaisesProhibitedRule.RULE_LABEL,
        identifier=None,
    )
    assert rule._check_occurrence_duplication(occurrences=[occurrence], filename="my_file.py", line_number=1) is False


def test_check_occurrence_duplication_no_duplication_via_line_no():
    rule = AssertRaisesProhibitedRule(file_path=Path("/path/to/file/my_file.py"), source_tree=ast.parse("a=1"))
    occurrence = Occurrence(
        line_number=1,
        filename="other_file.py",
        file_path=Path("/path/to/file/other_file.py"),
        rule_id=AssertRaisesProhibitedRule.RULE_ID,
        rule_label=AssertRaisesProhibitedRule.RULE_LABEL,
        identifier=None,
    )
    assert rule._check_occurrence_duplication(occurrences=[occurrence], filename="my_file.py", line_number=1) is False


def test_check_occurrence_duplication_has_duplication():
    rule = AssertRaisesProhibitedRule(file_path=Path("/path/to/file/my_file.py"), source_tree=ast.parse("a=1"))

    occurrence = Occurrence(
        line_number=7,
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        rule_id=AssertRaisesProhibitedRule.RULE_ID,
        rule_label=AssertRaisesProhibitedRule.RULE_LABEL,
        identifier=None,
    )
    assert rule._check_occurrence_duplication(occurrences=[occurrence], filename="my_file.py", line_number=7) is True


def test_check_direct_call_false():
    rule = AssertRaisesProhibitedRule(file_path=Path("/path/to/file/my_file.py"), source_tree=ast.parse("a=1"))

    code = "some_other_object.assertRaises(ValueError, func)"
    tree = ast.parse(code)
    node = tree.body[0].value

    assert rule._check_direct_call(node=node) is False


def test_check_context_manager_call_not_a_call():
    rule = AssertRaisesProhibitedRule(file_path=Path("/path/to/file/my_file.py"), source_tree=ast.parse("a=1"))

    code = """with obj.some_method as value:
         my_function()
     """
    tree = ast.parse(code)
    node = tree.body[0]

    assert rule._check_context_manager(node=node) is False


def test_check_context_manager_call_other_object():
    rule = AssertRaisesProhibitedRule(file_path=Path("/path/to/file/my_file.py"), source_tree=ast.parse("a=1"))

    code = """with some_object.assertRaises(ValueError):
        my_function()
     """
    tree = ast.parse(code)
    node = tree.body[0]

    assert rule._check_context_manager(node=node) is False


def test_assert_raises_in_context_manager():
    source_tree = ast.parse("""class MyTestCase(TestCase):
    def test_my_test(self):
        with self.assertRaises(RuntimeError):
            my_function()""")

    occurrences = AssertRaisesProhibitedRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=3,
        rule_id=AssertRaisesProhibitedRule.RULE_ID,
        rule_label=AssertRaisesProhibitedRule.RULE_LABEL,
        identifier=None,
    )


def test_assert_raises_direct_usage():
    source_tree = ast.parse("""class MyTestCase(TestCase):
    def test_my_test(self):
        self.assertRaises(RuntimeError)""")

    occurrences = AssertRaisesProhibitedRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=3,
        rule_id=AssertRaisesProhibitedRule.RULE_ID,
        rule_label=AssertRaisesProhibitedRule.RULE_LABEL,
        identifier=None,
    )


def test_assert_raises_message_used():
    source_tree = ast.parse("""class MyTestCase(TestCase):
    def test_my_test(self):
        with self.assertRaisesMessage(RuntimeError, "Hola mundo!"):
            my_function()""")

    occurrences = AssertRaisesProhibitedRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0



import ast

from boa_restrictor.common.rule import PYTHON_LINTING_RULE_PREFIX, Rule
from boa_restrictor.projections.occurrence import Occurrence


class ServiceClassHasOnlyOnePublicMethodRule(Rule):
    """
    Checks if service class, which is defined by ending on "Service" has only one public method called "process".
    """

    RULE_ID = f"{PYTHON_LINTING_RULE_PREFIX}005"
    RULE_LABEL = 'Service classes must have exactly one public method named "process".'

    def check(self) -> list[Occurrence]:
        occurrences = []

        for node in ast.walk(self.source_tree):
            if not isinstance(node, ast.ClassDef):
                continue
            if node.name.endswith("Service"):
                public_methods = []

                for item in node.body:
                    if isinstance(item, ast.FunctionDef):
                        # Check method names
                        if not item.name.startswith("_"):
                            public_methods.append(item.name)

                # Check whether the only public method is 'process'
                if public_methods != ["process"]:
                    occurrences.append(
                        Occurrence(
                            rule_id=self.RULE_ID,
                            rule_label=self.RULE_LABEL,
                            filename=self.filename,
                            file_path=self.file_path,
                            identifier=node.name,
                            line_number=node.lineno,
                        )
                    )

        return occurrences



import ast
from pathlib import Path
from unittest import mock

import pytest

from boa_restrictor.common.rule import Rule


@mock.patch.object(Rule, "check")
def test_run_check(mocked_check):
    Rule.run_check(file_path=Path("my/file.py"), source_tree=ast.parse("a=1"))

    mocked_check.assert_called_with()
    mocked_check.assert_called_once()


def test_init_variables_set():
    parsed_source_code = ast.parse("a=1")
    rule = Rule(file_path=Path("my/file.py"), source_tree=parsed_source_code)

    assert rule.file_path == Path("my/file.py")
    assert rule.filename == "file.py"
    assert rule.source_tree == parsed_source_code


def test_check_not_implemented():
    with pytest.raises(NotImplementedError):
        Rule.run_check(file_path=Path("my/file.py"), source_tree=ast.parse("a=1"))



from ambient_package_update.metadata.author import PackageAuthor
from ambient_package_update.metadata.constants import (
    DEV_DEPENDENCIES,
    LICENSE_MIT,
    SUPPORTED_DJANGO_VERSIONS,
    SUPPORTED_PYTHON_VERSIONS,
)
from ambient_package_update.metadata.executables import ScriptExecutable
from ambient_package_update.metadata.maintainer import PackageMaintainer
from ambient_package_update.metadata.package import PackageMetadata
from ambient_package_update.metadata.readme import ReadmeContent
from ambient_package_update.metadata.ruff_ignored_inspection import RuffIgnoredInspection

METADATA = PackageMetadata(
    package_name="boa-restrictor",
    github_package_group="ambient-innovation",
    authors=[
        PackageAuthor(
            name="Ambient Digital",
            email="hello@ambient.digital",
        ),
    ],
    maintainer=PackageMaintainer(name="Ambient Digital", url="https://ambient.digital/", email="hello@ambient.digital"),
    licenser="Ambient Innovation: GmbH",
    license=LICENSE_MIT,
    license_year=2024,
    development_status="4 - Beta",
    has_migrations=False,
    claim="A custom Python and Django linter from Ambient",
    readme_content=ReadmeContent(uses_internationalisation=False),
    main_branch="main",
    tests_require_django=False,
    dependencies=[f"Django>={SUPPORTED_DJANGO_VERSIONS[0]}", 'tomli; python_version < "3.11"'],
    supported_django_versions=SUPPORTED_DJANGO_VERSIONS,
    supported_python_versions=SUPPORTED_PYTHON_VERSIONS,
    optional_dependencies={
        "dev": [
            *DEV_DEPENDENCIES,
        ],
    },
    ruff_ignore_list=[
        RuffIgnoredInspection(key="TD002", comment="Missing issue link on the line following this TODO"),
        RuffIgnoredInspection(key="TD003", comment="Missing issue link on the line following this TODO"),
        RuffIgnoredInspection(key="PERF401", comment="Use `list.extend` to create a transformed list"),
    ],
    script_executables=[ScriptExecutable(name="boa-restrictor", import_path="boa_restrictor.cli.main:main")],
)



from boa_restrictor.common.noqa import get_noqa_comments


def test_get_noqa_comments_has_pbr_noqa():
    source_code = """x = 7  # noqa: PBR001"""

    result = get_noqa_comments(source_code=source_code)

    assert len(result) == 1
    assert isinstance(result[0], tuple)
    assert result[0][0] == 1
    assert result[0][1] == "# noqa: PBR001"


def test_get_noqa_comments_has_dbr_noqa():
    source_code = """from django.db.models import QuerySet  # noqa: DBR002"""

    result = get_noqa_comments(source_code=source_code)

    assert len(result) == 1
    assert isinstance(result[0], tuple)
    assert result[0][0] == 1
    assert result[0][1] == "# noqa: DBR002"


def test_get_noqa_comments_no_noqa_comment():
    source_code = """x = 7  # Great!"""

    result = get_noqa_comments(source_code=source_code)

    assert len(result) == 0



import ast
from pathlib import Path

from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules.django.no_assert_booleans_in_tests import ProhibitAssertBooleanInTests


def test_check_assert_true_found():
    source_tree = ast.parse("""class MyTest(TestCase):
    def test_x(self):
        self.assertTrue(x)""")

    occurrences = ProhibitAssertBooleanInTests.run_check(file_path=Path("test_x.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="test_x.py",
        file_path=Path("test_x.py"),
        line_number=3,
        rule_id=ProhibitAssertBooleanInTests.RULE_ID,
        rule_label=ProhibitAssertBooleanInTests.RULE_LABEL,
        identifier=None,
    )


def test_check_assert_false_found():
    source_tree = ast.parse("""class MyTest(TestCase):
    def test_x(self):
        self.assertFalse(x)""")

    occurrences = ProhibitAssertBooleanInTests.run_check(file_path=Path("test_x.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="test_x.py",
        file_path=Path("test_x.py"),
        line_number=3,
        rule_id=ProhibitAssertBooleanInTests.RULE_ID,
        rule_label=ProhibitAssertBooleanInTests.RULE_LABEL,
        identifier=None,
    )


def test_check_assert_via_test_case_class_indirect_usage():
    source_tree = ast.parse("""class MyTest(tests.TestCase):
    def test_x(self):
        self.assertFalse(x)""")

    occurrences = ProhibitAssertBooleanInTests.run_check(file_path=Path("test_x.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="test_x.py",
        file_path=Path("test_x.py"),
        line_number=3,
        rule_id=ProhibitAssertBooleanInTests.RULE_ID,
        rule_label=ProhibitAssertBooleanInTests.RULE_LABEL,
        identifier=None,
    )


def test_check_different_assert_not_found():
    source_tree = ast.parse("""class MyTest(TestCase):
    def test_x(self):
        self.assertIs(x, True)""")

    occurrences = ProhibitAssertBooleanInTests.run_check(file_path=Path("test_x.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_check_assert_on_non_self_object():
    source_tree = ast.parse("""class MyTest(TestCase):
    def test_x(self):
        some_other_object.assertTrue(x)""")

    occurrences = ProhibitAssertBooleanInTests.run_check(file_path=Path("test_x.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_check_assert_in_custom_testcase_subclass():
    """Verify the rule works with custom TestCase subclasses (not direct TestCase inheritance)."""
    source_tree = ast.parse("""class CustomTestCase(TestCase):
    pass

class MyTest(CustomTestCase):
    def test_x(self):
        self.assertTrue(x)""")

    occurrences = ProhibitAssertBooleanInTests.run_check(file_path=Path("test_x.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="test_x.py",
        file_path=Path("test_x.py"),
        line_number=6,
        rule_id=ProhibitAssertBooleanInTests.RULE_ID,
        rule_label=ProhibitAssertBooleanInTests.RULE_LABEL,
        identifier=None,
    )


def test_check_assert_with_class_variable():
    source_tree = ast.parse("""class MyTest(TestCase):
    something = 123""")

    occurrences = ProhibitAssertBooleanInTests.run_check(file_path=Path("test_x.py"), source_tree=source_tree)

    assert len(occurrences) == 0



import ast
from pathlib import Path

from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules import AvoidTypeHintsInVariableNamesAsSuffix


def test_check_variable_ends_on_list():
    source_tree = ast.parse("""user_list = []""")

    occurrences = AvoidTypeHintsInVariableNamesAsSuffix.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AvoidTypeHintsInVariableNamesAsSuffix.RULE_ID,
        rule_label=AvoidTypeHintsInVariableNamesAsSuffix.RULE_LABEL,
        identifier="user_list",
    )


def test_check_variable_ends_on_dict():
    source_tree = ast.parse("""user_dict = {}""")

    occurrences = AvoidTypeHintsInVariableNamesAsSuffix.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AvoidTypeHintsInVariableNamesAsSuffix.RULE_ID,
        rule_label=AvoidTypeHintsInVariableNamesAsSuffix.RULE_LABEL,
        identifier="user_dict",
    )


def test_check_variable_ends_on_set():
    source_tree = ast.parse("""user_set = ()""")

    occurrences = AvoidTypeHintsInVariableNamesAsSuffix.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AvoidTypeHintsInVariableNamesAsSuffix.RULE_ID,
        rule_label=AvoidTypeHintsInVariableNamesAsSuffix.RULE_LABEL,
        identifier="user_set",
    )


def test_check_variable_ends_on_str():
    source_tree = ast.parse("""user_str = ''""")

    occurrences = AvoidTypeHintsInVariableNamesAsSuffix.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AvoidTypeHintsInVariableNamesAsSuffix.RULE_ID,
        rule_label=AvoidTypeHintsInVariableNamesAsSuffix.RULE_LABEL,
        identifier="user_str",
    )


def test_check_variable_ends_on_int():
    source_tree = ast.parse("""user_int = 0""")

    occurrences = AvoidTypeHintsInVariableNamesAsSuffix.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AvoidTypeHintsInVariableNamesAsSuffix.RULE_ID,
        rule_label=AvoidTypeHintsInVariableNamesAsSuffix.RULE_LABEL,
        identifier="user_int",
    )


def test_check_variable_ends_on_float():
    source_tree = ast.parse("""user_float = 0.0""")

    occurrences = AvoidTypeHintsInVariableNamesAsSuffix.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AvoidTypeHintsInVariableNamesAsSuffix.RULE_ID,
        rule_label=AvoidTypeHintsInVariableNamesAsSuffix.RULE_LABEL,
        identifier="user_float",
    )


def test_check_variable_ends_on_bool():
    source_tree = ast.parse("""user_bool = False""")

    occurrences = AvoidTypeHintsInVariableNamesAsSuffix.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AvoidTypeHintsInVariableNamesAsSuffix.RULE_ID,
        rule_label=AvoidTypeHintsInVariableNamesAsSuffix.RULE_LABEL,
        identifier="user_bool",
    )


def test_check_variable_ends_on_qs():
    source_tree = ast.parse("""user_qs = UserQuerySet()""")

    occurrences = AvoidTypeHintsInVariableNamesAsSuffix.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AvoidTypeHintsInVariableNamesAsSuffix.RULE_ID,
        rule_label=AvoidTypeHintsInVariableNamesAsSuffix.RULE_LABEL,
        identifier="user_qs",
    )


def test_check_annotated_variable_ends_on_blocklisted_suffix():
    source_tree = ast.parse("""user_list: list = []""")

    occurrences = AvoidTypeHintsInVariableNamesAsSuffix.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AvoidTypeHintsInVariableNamesAsSuffix.RULE_ID,
        rule_label=AvoidTypeHintsInVariableNamesAsSuffix.RULE_LABEL,
        identifier="user_list",
    )


def test_check_annotated_variable_unpacking():
    source_tree = ast.parse("""my_dict[key]: str = 'value'""")

    occurrences = AvoidTypeHintsInVariableNamesAsSuffix.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0


def test_check_variable_no_specific_ending():
    source_tree = ast.parse("""users = []""")

    occurrences = AvoidTypeHintsInVariableNamesAsSuffix.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0


def test_check_variable_unpacking():
    source_tree = ast.parse("""a, b = [1, 2]""")

    occurrences = AvoidTypeHintsInVariableNamesAsSuffix.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0



import ast
from pathlib import Path

from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules import NoDjangoDbImportInViewsRule


def test_check_deep_import():
    source_tree = ast.parse("""from django.db.models.functions import Concat""")

    occurrences = NoDjangoDbImportInViewsRule.run_check(
        file_path=Path("/path/to/file/views.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="views.py",
        file_path=Path("/path/to/file/views.py"),
        line_number=1,
        rule_id=NoDjangoDbImportInViewsRule.RULE_ID,
        rule_label=NoDjangoDbImportInViewsRule.RULE_LABEL,
        identifier=None,
    )


def test_check_models_import():
    source_tree = ast.parse("""from django.db import models""")

    occurrences = NoDjangoDbImportInViewsRule.run_check(
        file_path=Path("/path/to/file/views.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="views.py",
        file_path=Path("/path/to/file/views.py"),
        line_number=1,
        rule_id=NoDjangoDbImportInViewsRule.RULE_ID,
        rule_label=NoDjangoDbImportInViewsRule.RULE_LABEL,
        identifier=None,
    )


def test_check_db_import():
    source_tree = ast.parse("""from django import db""")

    occurrences = NoDjangoDbImportInViewsRule.run_check(
        file_path=Path("/path/to/file/views.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="views.py",
        file_path=Path("/path/to/file/views.py"),
        line_number=1,
        rule_id=NoDjangoDbImportInViewsRule.RULE_ID,
        rule_label=NoDjangoDbImportInViewsRule.RULE_LABEL,
        identifier=None,
    )


def test_check_full_import():
    source_tree = ast.parse("""import django.db.models""")

    occurrences = NoDjangoDbImportInViewsRule.run_check(
        file_path=Path("/path/to/file/views.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="views.py",
        file_path=Path("/path/to/file/views.py"),
        line_number=1,
        rule_id=NoDjangoDbImportInViewsRule.RULE_ID,
        rule_label=NoDjangoDbImportInViewsRule.RULE_LABEL,
        identifier=None,
    )


def test_check_view_module():
    source_tree = ast.parse("""import django.db.models""")

    occurrences = NoDjangoDbImportInViewsRule.run_check(
        file_path=Path("/path/to/views/user.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="user.py",
        file_path=Path("/path/to/views/user.py"),
        line_number=1,
        rule_id=NoDjangoDbImportInViewsRule.RULE_ID,
        rule_label=NoDjangoDbImportInViewsRule.RULE_LABEL,
        identifier=None,
    )


def test_check_no_views_file():
    source_tree = ast.parse("""import django.db.models""")

    occurrences = NoDjangoDbImportInViewsRule.run_check(file_path=Path("managers.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_check_migrations_are_ok():
    source_tree = ast.parse("""import django.db.models.functions.comparison""")

    occurrences = NoDjangoDbImportInViewsRule.run_check(
        file_path=Path("migrations/0001_initial.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0


def test_check_no_db_import_import():
    source_tree = ast.parse("""import django.conf.settings""")

    occurrences = NoDjangoDbImportInViewsRule.run_check(
        file_path=Path("/path/to/file/views.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0


def test_check_no_db_form_import():
    source_tree = ast.parse("""from django.conf import settings""")

    occurrences = NoDjangoDbImportInViewsRule.run_check(
        file_path=Path("/path/to/file/views.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0


def test_check_relative_import():
    source_tree = ast.parse("""from . import db""")

    occurrences = NoDjangoDbImportInViewsRule.run_check(file_path=Path("/path/to/file/api.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_check_typing_type_hinting_imports_are_excluded():
    source_tree = ast.parse("""if typing.TYPE_CHECKING:
    from django.db import QuerySet""")

    occurrences = NoDjangoDbImportInViewsRule.run_check(
        file_path=Path("/path/to/file/views.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0


def test_check_constant_type_hinting_imports_are_excluded():
    source_tree = ast.parse("""if TYPE_CHECKING:
    from django.db import QuerySet""")

    occurrences = NoDjangoDbImportInViewsRule.run_check(
        file_path=Path("/path/to/file/views.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0


def test_check_other_ifs_are_ignored():
    source_tree = ast.parse("""if a == 1:
    from django.db import QuerySet""")

    occurrences = NoDjangoDbImportInViewsRule.run_check(
        file_path=Path("/path/to/file/views.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="views.py",
        file_path=Path("/path/to/file/views.py"),
        line_number=2,
        rule_id=NoDjangoDbImportInViewsRule.RULE_ID,
        rule_label=NoDjangoDbImportInViewsRule.RULE_LABEL,
        identifier=None,
    )



import ast

from boa_restrictor.common.rule import DJANGO_LINTING_RULE_PREFIX, Rule
from boa_restrictor.projections.occurrence import Occurrence


class AssertRaisesProhibitedRule(Rule):
    """
    Ensures that TestCase.assertRaises() is never used since asserting an exception without the actual error
    message leads to false positives.
    """

    RULE_ID = f"{DJANGO_LINTING_RULE_PREFIX}001"
    RULE_LABEL = 'Use of "assertRaises()" is discouraged. Use "assertRaisesMessage()" instead.'

    def _check_occurrence_duplication(self, *, occurrences: list[Occurrence], filename: str, line_number: int) -> bool:
        match_already_found = False
        for occurrence in occurrences:
            if occurrence.filename == filename and occurrence.line_number == line_number:
                match_already_found = True
        return match_already_found

    def _check_direct_call(self, node) -> bool:
        if isinstance(node.func, ast.Attribute):
            if node.func.attr == "assertRaises":
                if isinstance(node.func.value, ast.Name) and node.func.value.id == "self":
                    return True

        return False

    def _check_context_manager(self, node) -> bool:
        for item in node.items:
            if isinstance(item.context_expr, ast.Call):
                func = item.context_expr.func
                if isinstance(func, ast.Attribute) and func.attr == "assertRaises":
                    if isinstance(func.value, ast.Name) and func.value.id == "self":
                        return True
        return False

    def check(self) -> list[Occurrence]:
        occurrences = []

        for node in ast.walk(self.source_tree):
            node_matched = False

            # Direct call: self.assertRaises(...)
            if isinstance(node, ast.Call):
                node_matched = self._check_direct_call(node=node)

            # Context Manager: with self.assertRaises(...)
            elif isinstance(node, ast.With):
                node_matched = self._check_context_manager(node=node)

            if not node_matched:
                continue
            match_already_found = self._check_occurrence_duplication(
                occurrences=occurrences, filename=self.filename, line_number=node.lineno
            )

            if not match_already_found:
                occurrences.append(
                    Occurrence(
                        filename=self.filename,
                        file_path=self.file_path,
                        rule_label=self.RULE_LABEL,
                        rule_id=self.RULE_ID,
                        line_number=node.lineno,
                        identifier=None,
                    )
                )

        return occurrences



import ast

from boa_restrictor.common.rule import PYTHON_LINTING_RULE_PREFIX, Rule
from boa_restrictor.projections.occurrence import Occurrence


class AsteriskRequiredRule(Rule):
    """
    Checks if a method or function contains positional arguments (missing leading "*")
    "self" and "cls" are allowlisted, even if they are used outside a class context.
    """

    RULE_ID = f"{PYTHON_LINTING_RULE_PREFIX}001"
    RULE_LABEL = 'Positional arguments in functions and methods are discouraged. Add an "*" as the first argument.'

    def _missing_asterisk(self, *, node) -> bool:
        for arg in node.args.args:
            if isinstance(arg, ast.arg) and arg.arg not in ("self", "cls"):
                return True

        return False

    def check(self) -> list[Occurrence]:
        occurrences = []

        for node in ast.walk(self.source_tree):
            if not isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                continue
            if self._missing_asterisk(node=node):
                occurrences.append(
                    Occurrence(
                        rule_id=self.RULE_ID,
                        rule_label=self.RULE_LABEL,
                        filename=self.filename,
                        file_path=self.file_path,
                        identifier=node.name,
                        line_number=node.lineno,
                    )
                )

        return occurrences



import ast

from boa_restrictor.common.rule import DJANGO_LINTING_RULE_PREFIX, Rule
from boa_restrictor.projections.occurrence import Occurrence


class ProhibitDatetimeNow(Rule):
    """
    Prohibit the usage of "datetime.now()" in favour of "django.utils.timezone.now()".
    """

    RULE_ID = f"{DJANGO_LINTING_RULE_PREFIX}004"
    RULE_LABEL = 'Do not use "datetime.now()". Use "django.utils.timezone.now()" instead.'

    def check(self) -> list[Occurrence]:  # noqa: C901
        occurrences = []

        imports = {}
        for node in ast.walk(self.source_tree):
            # Collect import statements
            if isinstance(node, ast.Import):
                for alias in node.names:
                    imports[alias.asname or alias.name] = alias.name

            elif isinstance(node, ast.ImportFrom):
                module = node.module
                for alias in node.names:
                    full_name = f"{module}.{alias.name}" if module else alias.name
                    imports[alias.asname or alias.name] = full_name

        # Iterate source tree again to find all matches
        for node in ast.walk(self.source_tree):
            if isinstance(node, ast.Call):
                parts = []
                func = node.func

                # Reconstruct fully qualified name of the function call
                while isinstance(func, ast.Attribute):
                    parts.insert(0, func.attr)
                    func = func.value
                if isinstance(func, ast.Name):  # pragma: no cover (this "False" case seems to be invalid Python
                    parts.insert(0, func.id)
                    root = imports.get(parts[0], parts[0])
                    parts[0] = root

                full_call = ".".join(parts)

                if full_call in ("datetime.datetime.now", "datetime.now", "datetime.now"):
                    occurrences.append(
                        Occurrence(
                            filename=self.filename,
                            file_path=self.file_path,
                            rule_label=self.RULE_LABEL,
                            rule_id=self.RULE_ID,
                            line_number=node.lineno,
                            identifier=None,
                        )
                    )

        return occurrences



import ast

from boa_restrictor.common.rule import PYTHON_LINTING_RULE_PREFIX, Rule
from boa_restrictor.projections.occurrence import Occurrence


class ReturnStatementRequiresTypeHintRule(Rule):
    """
    This rule checks if a return statement is set, then a return type hint has to be defined.
    Doesn't match the case if there is no return statement but a type hint.
    """

    RULE_ID = f"{PYTHON_LINTING_RULE_PREFIX}002"
    RULE_LABEL = "Return statements require return type hint."

    def check(self) -> list[Occurrence]:
        occurrences = []

        for node in ast.walk(self.source_tree):
            if isinstance(node, ast.FunctionDef):
                has_return_statement = any(isinstance(child, ast.Return) for child in ast.walk(node))
                has_return_annotation = node.returns is not None

                if has_return_statement and not has_return_annotation:
                    occurrences.append(
                        Occurrence(
                            filename=self.filename,
                            file_path=self.file_path,
                            rule_label=self.RULE_LABEL,
                            rule_id=self.RULE_ID,
                            line_number=node.lineno,
                            identifier=node.name,
                        )
                    )

        return occurrences



import argparse
import ast
import os
import sys
from io import StringIO
from pathlib import Path
from unittest import mock

from boa_restrictor.cli.main import main
from boa_restrictor.common.rule import Rule
from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules import BOA_RESTRICTOR_RULES, DJANGO_BOA_RULES, AsteriskRequiredRule


@mock.patch.object(argparse.ArgumentParser, "parse_args")
def test_main_arguments_parsed(mocked_parse_args):
    main(
        argv=(
            "boa-restrictor",
            os.path.abspath(sys.argv[0]),
            "--config",
            "pyproject.toml",
        )
    )

    mocked_parse_args.assert_called_once_with(
        ("boa-restrictor", os.path.abspath(sys.argv[0]), "--config", "pyproject.toml"),
    )


@mock.patch("boa_restrictor.cli.main.load_configuration", return_value={"exclude": ["PBR001"]})
@mock.patch("builtins.open", mock.mock_open(read_data="# test file"))
@mock.patch.object(Rule, "run_check")
@mock.patch.object(AsteriskRequiredRule, "run_check")
def test_main_exclude_config_active(mocked_run_checks_asterisk, mocked_rule_run_checks, *args):
    main(
        argv=(
            os.path.abspath(sys.argv[0]),
            "--config",
            "pyproject.toml",
        )
    )

    mocked_run_checks_asterisk.assert_not_called()
    assert mocked_rule_run_checks.call_count == len(BOA_RESTRICTOR_RULES) + len(DJANGO_BOA_RULES) - 1, (
        "We expect all but one rule to be called."
    )


@mock.patch("boa_restrictor.cli.main.load_configuration", return_value={"per-file-excludes": {"*.py": ["PBR001"]}})
@mock.patch("builtins.open", mock.mock_open(read_data="# test file"))
@mock.patch.object(Rule, "run_check")
@mock.patch.object(AsteriskRequiredRule, "run_check")
def test_main_per_file_exclude_config_active(mocked_run_checks_asterisk, mocked_rule_run_checks, *args):
    main(
        argv=(
            "test_file.py",
            "--config",
            "pyproject.toml",
        )
    )

    mocked_run_checks_asterisk.assert_not_called()
    assert mocked_rule_run_checks.call_count == len(BOA_RESTRICTOR_RULES) + len(DJANGO_BOA_RULES) - 1, (
        "We expect all but one rule to be called."
    )


@mock.patch("boa_restrictor.cli.main.load_configuration", return_value={})
@mock.patch("builtins.open", mock.mock_open(read_data="# test file"))
@mock.patch("boa_restrictor.cli.main.get_rules")
def test_main_django_rules_default_enabled(mocked_get_rule, *args):
    main(
        argv=(
            os.path.abspath(sys.argv[0]),
            "--config",
            "pyproject.toml",
        )
    )

    mocked_get_rule.assert_called_with(use_django_rules=True)


@mock.patch("boa_restrictor.cli.main.load_configuration", return_value={"enable_django_rules": True})
@mock.patch("builtins.open", mock.mock_open(read_data="# test file"))
@mock.patch("boa_restrictor.cli.main.get_rules")
def test_main_django_rules_enabled(mocked_get_rule, *args):
    main(
        argv=(
            os.path.abspath(sys.argv[0]),
            "--config",
            "pyproject.toml",
        )
    )

    mocked_get_rule.assert_called_with(use_django_rules=True)


@mock.patch("boa_restrictor.cli.main.load_configuration", return_value={"enable_django_rules": False})
@mock.patch("builtins.open", mock.mock_open(read_data="# test file"))
@mock.patch("boa_restrictor.cli.main.get_rules")
def test_main_django_rules_disabled(mocked_get_rule, *args):
    main(
        argv=(
            os.path.abspath(sys.argv[0]),
            "--config",
            "pyproject.toml",
        )
    )

    mocked_get_rule.assert_called_with(use_django_rules=False)


def test_main_noqa_comments_called():
    with mock.patch("boa_restrictor.cli.main.get_noqa_comments", return_value=[]) as mocked_get_noqa_comments:
        with mock.patch("boa_restrictor.cli.main.load_configuration", return_value={}):
            with mock.patch("builtins.open", mock.mock_open(read_data="# test file")):
                main(
                    argv=(
                        os.path.abspath(sys.argv[0]),
                        "--config",
                        "pyproject.toml",
                    )
                )

                mocked_get_noqa_comments.assert_called_once()


@mock.patch.object(ast, "parse", side_effect=SyntaxError)
@mock.patch("boa_restrictor.cli.main.load_configuration", return_value={})
@mock.patch("builtins.open", mock.mock_open(read_data="# test file"))
def test_main_invalid_syntax(*args):
    # With syntax error in parsing, the function should raise the SyntaxError
    # since exception handling was removed in the revert
    try:
        main(
            argv=(
                os.path.abspath(sys.argv[0]),
                "--config",
                "pyproject.toml",
            )
        )
        # If no exception was raised, this test should fail
        raise AssertionError
    except SyntaxError:
        # This is expected behavior after the revert
        pass


def test_main_occurrences_are_written_to_cli():
    occurrence = Occurrence(
        rule_id="PBR000",
        rule_label="One to rule them all.",
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=42,
        identifier="my_function",
    )

    with mock.patch.object(sys.stdout, "write") as mocked_write:
        with mock.patch("boa_restrictor.cli.main.load_configuration", return_value={}):
            with mock.patch("builtins.open", mock.mock_open(read_data="# test file")):
                with mock.patch.object(
                    Rule,
                    "run_check",
                    return_value=[occurrence],
                ) as mocked_run_checks:
                    main(
                        argv=(
                            os.path.abspath(sys.argv[0]),
                            "--config",
                            "pyproject.toml",
                        )
                    )

                    # We have more than one rule
                    assert mocked_run_checks.call_count > 1

                    # We expect one line per occurrence
                    assert mocked_write.call_count == mocked_run_checks.call_count


@mock.patch("boa_restrictor.cli.main.load_configuration", return_value={})
@mock.patch("builtins.open", mock.mock_open(read_data="# test file"))
def test_main_occurrences_cli_output_correctly_formatted(mocked_load_config):
    occurrence = Occurrence(
        rule_id="PBR000",
        rule_label="One to rule them all.",
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=42,
        identifier="my_function",
    )

    with mock.patch("sys.stdout", new=StringIO()) as mock_stdout:
        with mock.patch.object(Rule, "run_check", return_value=[occurrence]):
            main(
                argv=(
                    os.path.abspath(sys.argv[0]),
                    "--config",
                    "pyproject.toml",
                )
            )
    actual_output = mock_stdout.getvalue()

    # Make test OS-independent
    actual_output = actual_output.replace("\\", "/")

    # Check that the formatting is correct
    assert '/path/to/file/my_file.py:42": ' in actual_output
    assert "(PBR000) One to rule them all." in actual_output


@mock.patch("boa_restrictor.cli.main.load_configuration", return_value={"per-file-excludes": {"*.py": ["PBR001"]}})
@mock.patch("builtins.open", mock.mock_open(read_data="# test file"))
@mock.patch("boa_restrictor.cli.main.is_rule_excluded_per_file", return_value=True)
@mock.patch.object(AsteriskRequiredRule, "run_check")
def test_main_per_file_exclusion_skips_rule(mocked_run_check, mocked_is_excluded, *args):
    """Test that line 59 continue statement is executed when per-file exclusion matches"""
    main(
        argv=(
            os.path.abspath(sys.argv[0]),
            "--config",
            "pyproject.toml",
        )
    )

    # Verify the rule was excluded via per-file config
    mocked_is_excluded.assert_called()
    # Verify that the rule check was NOT called for the excluded rule
    mocked_run_check.assert_not_called()



import ast
from pathlib import Path

from boa_restrictor.common.rule import PYTHON_LINTING_RULE_PREFIX, Rule
from boa_restrictor.projections.occurrence import Occurrence


class NoLoopsInTestsRule(Rule):
    """
    Prohibits loops in tests since tests should be as simple as possible.
    """

    RULE_ID = f"{PYTHON_LINTING_RULE_PREFIX}008"
    RULE_LABEL = "Using loops in unit-tests is discouraged."

    def check(self) -> list[Occurrence]:
        occurrences = []

        if not self._is_test_file(self.file_path):
            return occurrences

        for node in ast.walk(self.source_tree):
            if isinstance(node, ast.FunctionDef) and node.name.startswith("test"):
                if self._contains_loop_or_comprehension(node):
                    occurrences.append(
                        Occurrence(
                            rule_id=self.RULE_ID,
                            rule_label=self.RULE_LABEL,
                            filename=self.filename,
                            file_path=self.file_path,
                            identifier=node.name,
                            line_number=node.lineno,
                        )
                    )

        return occurrences

    @staticmethod
    def _is_test_file(filepath: Path) -> bool:
        return (
            any(part == "tests" for part in filepath.parts)
            and filepath.name.startswith("test_")
            and filepath.name.endswith(".py")
        )

    def _contains_loop_or_comprehension(self, node) -> bool:
        for child in ast.iter_child_nodes(node):
            if isinstance(child, (ast.For, ast.While, ast.ListComp, ast.SetComp, ast.DictComp)):
                return True
            if self._contains_loop_or_comprehension(child):
                return True
        return False



from boa_restrictor.rules.django.avoid_tuple_based_model_choices import AvoidTupleBasedModelChoices
from boa_restrictor.rules.django.no_assert_booleans_in_tests import ProhibitAssertBooleanInTests
from boa_restrictor.rules.django.no_db_in_api import NoDjangoDbImportInApiRule
from boa_restrictor.rules.django.no_db_in_views import NoDjangoDbImportInViewsRule
from boa_restrictor.rules.django.prohibit_assert_raises import AssertRaisesProhibitedRule
from boa_restrictor.rules.django.prohibit_datetime_now import ProhibitDatetimeNow
from boa_restrictor.rules.python.abstract_class_inherits_from_abc import AbstractClassesInheritFromAbcRule
from boa_restrictor.rules.python.asterisk_required import AsteriskRequiredRule
from boa_restrictor.rules.python.dataclass_kw_only import DataclassWithKwargsOnlyRule
from boa_restrictor.rules.python.global_import_datetime import GlobalImportDatetimeRule
from boa_restrictor.rules.python.no_loops_in_tests import NoLoopsInTestsRule
from boa_restrictor.rules.python.no_type_hints_in_variable_names import AvoidTypeHintsInVariableNamesAsSuffix
from boa_restrictor.rules.python.return_type_hints import ReturnStatementRequiresTypeHintRule
from boa_restrictor.rules.python.service_class_only_one_public import ServiceClassHasOnlyOnePublicMethodRule

BOA_RESTRICTOR_RULES = (
    AsteriskRequiredRule,
    ReturnStatementRequiresTypeHintRule,
    GlobalImportDatetimeRule,
    DataclassWithKwargsOnlyRule,
    ServiceClassHasOnlyOnePublicMethodRule,
    AbstractClassesInheritFromAbcRule,
    AvoidTypeHintsInVariableNamesAsSuffix,
    NoLoopsInTestsRule,
)

DJANGO_BOA_RULES = (
    AssertRaisesProhibitedRule,
    NoDjangoDbImportInViewsRule,
    ProhibitAssertBooleanInTests,
    ProhibitDatetimeNow,
    NoDjangoDbImportInApiRule,
    AvoidTupleBasedModelChoices,
)


def get_rules(*, use_django_rules: bool) -> tuple:
    """
    Returns a list of all enabled rules.
    """
    if use_django_rules:
        return BOA_RESTRICTOR_RULES + DJANGO_BOA_RULES
    return BOA_RESTRICTOR_RULES



import ast

from boa_restrictor.common.rule import DJANGO_LINTING_RULE_PREFIX, Rule
from boa_restrictor.projections.occurrence import Occurrence


class ProhibitAssertBooleanInTests(Rule):
    """
    Prohibit the usage of "assertTrue" and "assertFalse" in Django unittests.
    """

    RULE_ID = f"{DJANGO_LINTING_RULE_PREFIX}003"
    RULE_LABEL = (
        'Do not use "assertTrue" or "assertFalse" in Django unittests. Use "assertIs(x, True)" or '
        '"assertIs(x, False)" instead.'
    )

    def check(self) -> list[Occurrence]:
        occurrences = []
        for node in ast.walk(self.source_tree):
            if not isinstance(node, ast.Call):
                continue

            func = node.func
            if isinstance(func, ast.Attribute) and func.attr in {"assertTrue", "assertFalse"}:
                if isinstance(func.value, ast.Name) and func.value.id == "self":
                    occurrences.append(
                        Occurrence(
                            filename=self.filename,
                            file_path=self.file_path,
                            rule_label=self.RULE_LABEL,
                            rule_id=self.RULE_ID,
                            line_number=node.lineno,
                            identifier=None,
                        )
                    )

        return occurrences



import ast
from pathlib import Path

from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules.python.abstract_class_inherits_from_abc import AbstractClassesInheritFromAbcRule


def test_functions_named_abstract_are_ignored():
    source_tree = ast.parse("""def abstract_function():
            pass""")

    occurrences = AbstractClassesInheritFromAbcRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0


def test_classes_not_named_abstract_are_ignored():
    source_tree = ast.parse("""class BeerHeuristics:
            pass""")

    occurrences = AbstractClassesInheritFromAbcRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0


def test_abstract_class_missing_abc_inheritance():
    source_tree = ast.parse("""class AbstractService:
            pass""")

    occurrences = AbstractClassesInheritFromAbcRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AbstractClassesInheritFromAbcRule.RULE_ID,
        rule_label=AbstractClassesInheritFromAbcRule.RULE_LABEL,
        identifier="AbstractService",
    )


def test_abstract_class_missing_abc_inheritance_but_has_other():
    source_tree = ast.parse("""class AbstractService(BaseClass):
            pass""")

    occurrences = AbstractClassesInheritFromAbcRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AbstractClassesInheritFromAbcRule.RULE_ID,
        rule_label=AbstractClassesInheritFromAbcRule.RULE_LABEL,
        identifier="AbstractService",
    )


def test_abstract_class_missing_abc_odd_inhertiance():
    source_tree = ast.parse("""class AbstractService(dynamic_base()):
            pass""")

    occurrences = AbstractClassesInheritFromAbcRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AbstractClassesInheritFromAbcRule.RULE_ID,
        rule_label=AbstractClassesInheritFromAbcRule.RULE_LABEL,
        identifier="AbstractService",
    )


def test_abstract_class_having_abc_inheritance():
    source_tree = ast.parse("""class AbstractService(abc.ABC):
            pass""")

    occurrences = AbstractClassesInheritFromAbcRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0


def test_abstract_class_having_direct_abc_inheritance():
    source_tree = ast.parse("""class AbstractService(ABC):
            pass""")

    occurrences = AbstractClassesInheritFromAbcRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0


def test_abstract_class_having_abc_and_other_inheritance():
    source_tree = ast.parse("""class AbstractService(MyMixin, abc.ABC):
            pass""")

    occurrences = AbstractClassesInheritFromAbcRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0


def test_abstract_class_via_metaclass():
    source_tree = ast.parse("""class MyAbstractMetaClass(metaclass=ABCMeta):
    pass""")

    occurrences = AbstractClassesInheritFromAbcRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0


def test_abstract_class_via_abc_metaclass():
    source_tree = ast.parse("""class MyAbstractMetaClass(metaclass=abc.ABCMeta):
    pass""")

    occurrences = AbstractClassesInheritFromAbcRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 0


def test_abstract_class_base_kwarg_not_metaclass():
    source_tree = ast.parse("""class MyAbstractMetaClass(test=True):
    pass""")

    occurrences = AbstractClassesInheritFromAbcRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AbstractClassesInheritFromAbcRule.RULE_ID,
        rule_label=AbstractClassesInheritFromAbcRule.RULE_LABEL,
        identifier="MyAbstractMetaClass",
    )


def test_abstract_class_base_kwarg_dynamic():
    source_tree = ast.parse("""class MyAbstractMetaClass(metaclass=dynamic_metaclass()):
    pass""")

    occurrences = AbstractClassesInheritFromAbcRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AbstractClassesInheritFromAbcRule.RULE_ID,
        rule_label=AbstractClassesInheritFromAbcRule.RULE_LABEL,
        identifier="MyAbstractMetaClass",
    )


def test_abstract_class_lower_case():
    source_tree = ast.parse("""class abstractService:
    pass""")

    occurrences = AbstractClassesInheritFromAbcRule.run_check(
        file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree
    )

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AbstractClassesInheritFromAbcRule.RULE_ID,
        rule_label=AbstractClassesInheritFromAbcRule.RULE_LABEL,
        identifier="abstractService",
    )



import re
import tokenize
from io import StringIO

from boa_restrictor.common.rule import DJANGO_LINTING_RULE_PREFIX, PYTHON_LINTING_RULE_PREFIX


def get_noqa_comments(*, source_code: str) -> list[tuple[int, str]]:
    """
    Walk the target code and detect all lines having a "# noqa" comment with "our" prefix.
    """
    noqa_statements = []

    tokens = tokenize.generate_tokens(StringIO(source_code).readline)
    patterns = [
        re.compile(r"^#\snoqa:\s*.*?" + PYTHON_LINTING_RULE_PREFIX + r"\d{3}"),
        re.compile(r"^#\snoqa:\s*.*?" + DJANGO_LINTING_RULE_PREFIX + r"\d{3}"),
    ]

    for token in tokens:
        token_type, token_string, start, _, _ = token

        for pattern in patterns:
            if token_type == tokenize.COMMENT and pattern.search(token_string):
                noqa_statements.append((start[0], token_string.strip()))

    return noqa_statements



import ast

from boa_restrictor.common.rule import PYTHON_LINTING_RULE_PREFIX, Rule
from boa_restrictor.projections.occurrence import Occurrence


class AvoidTypeHintsInVariableNamesAsSuffix(Rule):
    """
    This rule will enforce that variables don't contain type hints as suffixes like "user_list" or "project_qs".
    """

    RULE_ID = f"{PYTHON_LINTING_RULE_PREFIX}007"
    RULE_LABEL = "Prohibit type-hinting in variable names"

    BAD_SUFFIXES = ("_list", "_dict", "_set", "_str", "_int", "_float", "_bool", "_qs")

    def check(self) -> list[Occurrence]:  # noqa: C901
        occurrences = []

        for node in ast.walk(self.source_tree):
            if isinstance(node, ast.Assign):
                for target in node.targets:
                    if not isinstance(target, ast.Name):
                        continue
                    var_name = target.id
                    for suffix in self.BAD_SUFFIXES:
                        if var_name.endswith(suffix):
                            occurrences.append(
                                Occurrence(
                                    rule_id=self.RULE_ID,
                                    rule_label=self.RULE_LABEL,
                                    filename=self.filename,
                                    file_path=self.file_path,
                                    identifier=var_name,
                                    line_number=node.lineno,
                                )
                            )
            # For type annotations
            elif isinstance(node, ast.AnnAssign):
                if not isinstance(node.target, ast.Name):
                    continue
                var_name = node.target.id
                for suffix in self.BAD_SUFFIXES:
                    if var_name.endswith(suffix):
                        occurrences.append(
                            Occurrence(
                                rule_id=self.RULE_ID,
                                rule_label=self.RULE_LABEL,
                                filename=self.filename,
                                file_path=self.file_path,
                                identifier=var_name,
                                line_number=node.lineno,
                            )
                        )

        return occurrences



import pytest

from boa_restrictor.exceptions.configuration import TomlParsingError


def test_toml_syntax_error():
    exception = TomlParsingError(filename="my_file")

    with pytest.raises(TomlParsingError, match=r'TOML file "my_file" contains syntax errors.'):
        raise exception



import ast
from pathlib import Path

from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules.django.avoid_tuple_based_model_choices import AvoidTupleBasedModelChoices


def test_tuple_based_choices_in_model_found():
    source_tree = ast.parse("""class MyModel(models.Model):
    STATUS_CHOICES = (
        ('A', 'Active'),
        ('I', 'Inactive'),
    )
    status = models.CharField(max_length=1, choices=STATUS_CHOICES)""")

    occurrences = AvoidTupleBasedModelChoices.run_check(file_path=Path("/path/to/file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="file.py",
        file_path=Path("/path/to/file.py"),
        line_number=2,
        rule_id=AvoidTupleBasedModelChoices.RULE_ID,
        rule_label=AvoidTupleBasedModelChoices.RULE_LABEL,
        identifier=None,
    )


def test_tuple_based_choices_outside_of_model_found():
    source_tree = ast.parse("""STATUS_CHOICES = (
    ('A', 'Active'),
    ('I', 'Inactive'),
)

class MyModel(models.Model):
    status = models.CharField(max_length=1, choices=STATUS)""")

    occurrences = AvoidTupleBasedModelChoices.run_check(file_path=Path("/path/to/file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="file.py",
        file_path=Path("/path/to/file.py"),
        line_number=2,
        rule_id=AvoidTupleBasedModelChoices.RULE_ID,
        rule_label=AvoidTupleBasedModelChoices.RULE_LABEL,
        identifier=None,
    )


def test_integer_choices_ok():
    source_tree = ast.parse("""class MyModel(models.Model):
    class StatusChoices(models.IntegerChoices):
        ACTIVE = 1, "Active"
        INACTIVE = 2, "Inactive"
        PENDING = 3, "Pending"

    status = models.IntegerField(choices=StatusChoices.choices)""")

    occurrences = AvoidTupleBasedModelChoices.run_check(file_path=Path("/path/to/file.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_string_choices_ok():
    source_tree = ast.parse("""class MyModel(models.Model):
    class StatusChoices(models.StringChoices):
        ACTIVE = "active", "Active"
        INACTIVE = "inactive", "Inactive"
        PENDING = "pending", "Pending"

    status = models.CharField(choices=StatusChoices.choices, max_length=20)""")

    occurrences = AvoidTupleBasedModelChoices.run_check(file_path=Path("/path/to/file.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_direct_model_inheritance_found():
    """Test line 34: Direct Model inheritance (not models.Model)"""
    source_tree = ast.parse("""class MyModel(Model):
    STATUS_CHOICES = (
        ('A', 'Active'),
        ('I', 'Inactive'),
    )
    status = CharField(max_length=1, choices=STATUS_CHOICES)""")

    occurrences = AvoidTupleBasedModelChoices.run_check(file_path=Path("/path/to/file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0].line_number == 2  # noqa: PLR2004


def test_empty_tuple_not_detected():
    """Test line 46: Empty tuple/list structures should not be detected"""
    source_tree = ast.parse("""class MyModel(models.Model):
    STATUS_CHOICES = ()
    EMPTY_LIST = []
    status = models.CharField(max_length=1)""")

    occurrences = AvoidTupleBasedModelChoices.run_check(file_path=Path("/path/to/file.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_tuple_assignment_without_choices_name_not_detected():
    """Test line 91->90: Tuple-based assignment that doesn't end with 'CHOICES' should not be detected"""
    source_tree = ast.parse("""SOME_TUPLES = (
    ('A', 'Active'),
    ('I', 'Inactive'),
)

class MyModel(models.Model):
    status = models.CharField(max_length=1)""")

    occurrences = AvoidTupleBasedModelChoices.run_check(file_path=Path("/path/to/file.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_invalid_tuple_structure_not_detected():
    """Test case where tuple elements are not 2-tuples"""
    source_tree = ast.parse("""class MyModel(models.Model):
    INVALID_CHOICES = (
        ('A',),  # Only one element
        ('B', 'Beta', 'Extra'),  # Three elements
    )
    status = models.CharField(max_length=1)""")

    occurrences = AvoidTupleBasedModelChoices.run_check(file_path=Path("/path/to/file.py"), source_tree=source_tree)

    assert len(occurrences) == 0



import ast
from pathlib import Path

from boa_restrictor.common.rule import DJANGO_LINTING_RULE_PREFIX, Rule
from boa_restrictor.projections.occurrence import Occurrence


class NoDjangoDbImportInApiRule(Rule):
    """
    Ensures that no Django low-level database functionality is imported and therefore used in the API layer.
    """

    RULE_ID = f"{DJANGO_LINTING_RULE_PREFIX}005"
    RULE_LABEL = 'Do not use "django.db" in the API layer. Move it to a manager instead.'

    def is_api_file(self, path: Path) -> bool:
        path = path.resolve()

        if path.name == "api.py":
            return True

        return "api" in path.parts

    def is_type_checking_if(self, node) -> bool:
        # if TYPE_CHECKING:
        if isinstance(node, ast.If):
            test = node.test
            # Case 1: TYPE_CHECKING
            if isinstance(test, ast.Name) and test.id == "TYPE_CHECKING":
                return True
            # Case 2: typing.TYPE_CHECKING
            elif (
                isinstance(test, ast.Attribute)
                and isinstance(test.value, ast.Name)
                and test.value.id == "typing"
                and test.attr == "TYPE_CHECKING"
            ):
                return True
        return False

    def check(self) -> list[Occurrence]:  # noqa: C901
        occurrences = []
        type_checking_lines = set()

        if not self.is_api_file(path=self.file_path):
            return occurrences

        # Find and store all type-checking imports by line number
        for node in ast.walk(self.source_tree):
            if self.is_type_checking_if(node):
                for inner in node.body:
                    for subnode in ast.walk(inner):
                        if isinstance(subnode, (ast.Import, ast.ImportFrom)):
                            type_checking_lines.add(subnode.lineno)

        for node in ast.walk(self.source_tree):
            if isinstance(node, ast.Import):
                for alias in node.names:
                    if node.lineno not in type_checking_lines and alias.name.startswith("django.db"):
                        occurrences.append(
                            Occurrence(
                                filename=self.filename,
                                file_path=self.file_path,
                                rule_label=self.RULE_LABEL,
                                rule_id=self.RULE_ID,
                                line_number=node.lineno,
                                identifier=None,
                            )
                        )
            elif isinstance(node, ast.ImportFrom):
                if node.lineno not in type_checking_lines and (
                    (node.module and node.module.startswith("django.db"))
                    or (node.module == "django" and any(alias.name == "db" for alias in node.names))
                ):
                    occurrences.append(
                        Occurrence(
                            filename=self.filename,
                            file_path=self.file_path,
                            rule_label=self.RULE_LABEL,
                            rule_id=self.RULE_ID,
                            line_number=node.lineno,
                            identifier=None,
                        )
                    )

        return occurrences



import ast

from boa_restrictor.common.rule import DJANGO_LINTING_RULE_PREFIX, Rule
from boa_restrictor.projections.occurrence import Occurrence


class AvoidTupleBasedModelChoices(Rule):
    """
    Prohibit the usage of tuple-based choices in model fields.
    Use new class-based choices instead.
    """

    # Constant for tuple-based choice validation
    CHOICE_TUPLE_LENGTH = 2

    RULE_ID = f"{DJANGO_LINTING_RULE_PREFIX}006"
    RULE_LABEL = "Avoid using old tuple-based Django model choices. Use class-based choices instead."

    def _is_django_model(self, node: ast.ClassDef) -> bool:
        """
        Check if a class inherits from models.Model (directly)
        """
        for base in node.bases:
            # models.Model
            if (
                isinstance(base, ast.Attribute)
                and base.attr == "Model"
                and isinstance(base.value, ast.Name)
                and base.value.id == "models"
            ):
                return True
            # or direct Model
            if isinstance(base, ast.Name) and base.id == "Model":
                return True
        return False

    def _is_tuple_based_choices(self, value: ast.AST) -> bool:
        """
        Detection of old tuple-based choices:
        - Either ast.Tuple / ast.List
        - Elements must each be ast.Tuple (e.g. ('A','Active'))
        """
        if isinstance(value, (ast.Tuple, ast.List)):
            # Ignore empty structures
            if not value.elts:
                return False
            # Check if each element is a 2-tuple
            for elt in value.elts:
                if not (isinstance(elt, ast.Tuple) and len(elt.elts) == self.CHOICE_TUPLE_LENGTH):
                    return False
            return True
        return False

    def _is_choices_variable_name(self, target_name: str) -> bool:
        """
        Check if variable name suggests it contains choices (ends with 'CHOICES')
        """
        return target_name.upper().endswith("CHOICES")

    def _create_occurrence(self, line_number: int) -> Occurrence:
        """Create an occurrence for a tuple-based choices violation."""
        return Occurrence(
            filename=self.filename,
            file_path=self.file_path,
            rule_label=self.RULE_LABEL,
            rule_id=self.RULE_ID,
            line_number=line_number,
            identifier=None,
        )

    def _check_django_model_assignments(self, occurrences: list[Occurrence]) -> set[int]:
        """Check assignments inside Django model classes and return set of processed assignment IDs."""
        django_model_assignments = set()

        for node in ast.walk(self.source_tree):
            if isinstance(node, ast.ClassDef) and self._is_django_model(node):
                for stmt in node.body:
                    if isinstance(stmt, ast.Assign):
                        django_model_assignments.add(id(stmt))
                        if self._is_tuple_based_choices(stmt.value):
                            occurrences.append(self._create_occurrence(stmt.lineno))

        return django_model_assignments

    def _check_standalone_assignments(self, occurrences: list[Occurrence], django_model_assignments: set[int]) -> None:
        """Check tuple-based choices assignments outside of Django models."""
        for node in ast.walk(self.source_tree):
            if isinstance(node, ast.Assign) and id(node) not in django_model_assignments:
                if self._is_tuple_based_choices(node.value):
                    for target in node.targets:
                        if isinstance(target, ast.Name) and self._is_choices_variable_name(target.id):
                            # Report the line number of the first tuple element for better user experience
                            line_number = node.value.elts[0].lineno if node.value.elts else node.lineno
                            occurrences.append(self._create_occurrence(line_number))

    def check(self) -> list[Occurrence]:
        occurrences: list[Occurrence] = []

        # Check assignments inside Django models first
        django_model_assignments = self._check_django_model_assignments(occurrences)

        # Check standalone assignments outside of models
        self._check_standalone_assignments(occurrences, django_model_assignments)

        return occurrences



import ast
from pathlib import Path

from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules.python.asterisk_required import AsteriskRequiredRule


def test_method_has_asterisk():
    source_tree = ast.parse("""class MyClass:
        def my_method(self, *, a):
            pass
    """)

    occurrences = AsteriskRequiredRule.run_check(file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_method_no_params():
    source_tree = ast.parse("""class MyClass:
        def my_method(self):
            pass
    """)

    occurrences = AsteriskRequiredRule.run_check(file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_method_missing_asterisk():
    source_tree = ast.parse("""class MyClass:
        def my_method(self, a):
            pass
    """)

    occurrences = AsteriskRequiredRule.run_check(file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=2,
        rule_id=AsteriskRequiredRule.RULE_ID,
        rule_label=AsteriskRequiredRule.RULE_LABEL,
        identifier="my_method",
    )


def test_function_has_asterisk():
    source_tree = ast.parse("""def my_function(*, a):
        pass
    """)

    occurrences = AsteriskRequiredRule.run_check(file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_function_no_params():
    source_tree = ast.parse("""def my_function():
        pass
    """)

    occurrences = AsteriskRequiredRule.run_check(file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_function_missing_asterisk():
    source_tree = ast.parse("""def my_function(a):
        pass
    """)

    occurrences = AsteriskRequiredRule.run_check(file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AsteriskRequiredRule.RULE_ID,
        rule_label=AsteriskRequiredRule.RULE_LABEL,
        identifier="my_function",
    )


def test_function_asterisk_too_late():
    source_tree = ast.parse("""def my_function(a, *, b):
        pass
    """)

    occurrences = AsteriskRequiredRule.run_check(file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AsteriskRequiredRule.RULE_ID,
        rule_label=AsteriskRequiredRule.RULE_LABEL,
        identifier="my_function",
    )


def test_function_arg_with_defaults():
    source_tree = ast.parse("""def my_function(a=42):
        pass
    """)

    occurrences = AsteriskRequiredRule.run_check(file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AsteriskRequiredRule.RULE_ID,
        rule_label=AsteriskRequiredRule.RULE_LABEL,
        identifier="my_function",
    )


def test_async_function_has_asterisk():
    source_tree = ast.parse("""async def my_function(*, a):
        pass
    """)

    occurrences = AsteriskRequiredRule.run_check(file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_async_function_missing_asterisk():
    source_tree = ast.parse("""async def my_function(a):
        pass
    """)

    occurrences = AsteriskRequiredRule.run_check(file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AsteriskRequiredRule.RULE_ID,
        rule_label=AsteriskRequiredRule.RULE_LABEL,
        identifier="my_function",
    )


def test_self_outside_of_class_not_matched():
    source_tree = ast.parse("""def my_function(self):
        pass
    """)

    occurrences = AsteriskRequiredRule.run_check(file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_cls_outside_of_class_not_matched():
    source_tree = ast.parse("""def my_function(cls):
        pass
    """)

    occurrences = AsteriskRequiredRule.run_check(file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 0


def test_leading_cls_and_following_attributes():
    source_tree = ast.parse("""def my_function(cls, a):
        pass
    """)

    occurrences = AsteriskRequiredRule.run_check(file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="my_file.py",
        file_path=Path("/path/to/file/my_file.py"),
        line_number=1,
        rule_id=AsteriskRequiredRule.RULE_ID,
        rule_label=AsteriskRequiredRule.RULE_LABEL,
        identifier="my_function",
    )


def test_lambda_not_matched():
    source_tree = ast.parse("""double = lambda x: x * 2""")

    occurrences = AsteriskRequiredRule.run_check(file_path=Path("/path/to/file/my_file.py"), source_tree=source_tree)

    assert len(occurrences) == 0



from pathlib import Path

from boa_restrictor.rules import BOA_RESTRICTOR_RULES, DJANGO_BOA_RULES, get_rules


def test_boa_restrictor_rules_constant_not_missing_rules():
    """
    This test is a check to ensure we don't forget to register new rules.
    """
    number_of_rule_files = 0

    # Python rules
    for file in (Path(__file__).resolve().parent.parent.parent / "boa_restrictor/rules/python").iterdir():
        if file.suffix == ".py" and file.name != "__init__.py":
            number_of_rule_files += 1

    # Django rules
    for file in (Path(__file__).resolve().parent.parent.parent / "boa_restrictor/rules/django").iterdir():
        if file.suffix == ".py" and file.name != "__init__.py":
            number_of_rule_files += 1

    assert len(get_rules(use_django_rules=True)) == number_of_rule_files


def test_get_rules_django_rules_enabled():
    assert get_rules(use_django_rules=True) == BOA_RESTRICTOR_RULES + DJANGO_BOA_RULES


def test_get_rules_django_rules_disabled():
    assert get_rules(use_django_rules=False) == BOA_RESTRICTOR_RULES



import ast

from boa_restrictor.common.rule import PYTHON_LINTING_RULE_PREFIX, Rule
from boa_restrictor.projections.occurrence import Occurrence


class GlobalImportDatetimeRule(Rule):
    """
    Checks if the "datetime" object was imported from the "datetime" module.
    This will lead to unclear and inconsistent code. Thus, we enforce a single way of using the datetime module.
    """

    RULE_ID = f"{PYTHON_LINTING_RULE_PREFIX}003"
    RULE_LABEL = "Prohibiting nested import of datetime from datetime module."

    def check(self) -> list[Occurrence]:
        occurrences = []

        for node in ast.walk(self.source_tree):
            if isinstance(node, ast.ImportFrom) and node.module == "datetime":
                for alias in node.names:
                    if alias.name in {"datetime", "date"}:
                        occurrences.append(
                            Occurrence(
                                filename=self.filename,
                                file_path=self.file_path,
                                rule_label=self.RULE_LABEL,
                                rule_id=self.RULE_ID,
                                line_number=node.lineno,
                                identifier=None,
                            )
                        )

        return occurrences



import ast
from pathlib import Path

from boa_restrictor.projections.occurrence import Occurrence
from boa_restrictor.rules.django.prohibit_datetime_now import ProhibitDatetimeNow


def test_datetime_now_import_found():
    source_tree = ast.parse("""from datetime import datetime
timestamp = datetime.now()""")

    occurrences = ProhibitDatetimeNow.run_check(file_path=Path("/path/to/file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="file.py",
        file_path=Path("/path/to/file.py"),
        line_number=2,
        rule_id=ProhibitDatetimeNow.RULE_ID,
        rule_label=ProhibitDatetimeNow.RULE_LABEL,
        identifier=None,
    )


def test_direct_import_found():
    source_tree = ast.parse("""import datetime
timestamp = datetime.datetime.now()""")

    occurrences = ProhibitDatetimeNow.run_check(file_path=Path("/path/to/file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="file.py",
        file_path=Path("/path/to/file.py"),
        line_number=2,
        rule_id=ProhibitDatetimeNow.RULE_ID,
        rule_label=ProhibitDatetimeNow.RULE_LABEL,
        identifier=None,
    )


def test_sub_import_found():
    source_tree = ast.parse("""from datetime.datetime import now
timestamp = now()""")

    occurrences = ProhibitDatetimeNow.run_check(file_path=Path("/path/to/file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="file.py",
        file_path=Path("/path/to/file.py"),
        line_number=2,
        rule_id=ProhibitDatetimeNow.RULE_ID,
        rule_label=ProhibitDatetimeNow.RULE_LABEL,
        identifier=None,
    )


def test_renamed_import_found():
    source_tree = ast.parse("""from datetime import datetime as dt
timestamp = dt.now()""")

    occurrences = ProhibitDatetimeNow.run_check(file_path=Path("/path/to/file.py"), source_tree=source_tree)

    assert len(occurrences) == 1
    assert occurrences[0] == Occurrence(
        filename="file.py",
        file_path=Path("/path/to/file.py"),
        line_number=2,
        rule_id=ProhibitDatetimeNow.RULE_ID,
        rule_label=ProhibitDatetimeNow.RULE_LABEL,
        identifier=None,
    )


def test_other_datetime_import_not_found():
    source_tree = ast.parse("""from datetime import datetime
timestamp = datetime.today()""")

    occurrences = ProhibitDatetimeNow.run_check(file_path=Path("/path/to/file.py"), source_tree=source_tree)

    assert len(occurrences) == 0



import dataclasses
from pathlib import Path


# TODO: Add kw_only when we drop Python 3.9 support
@dataclasses.dataclass
class Occurrence:
    rule_id: str
    rule_label: str
    filename: str
    file_path: Path
    identifier: str | None
    line_number: int
